@@ -171,9 +171,15 @@ class _OptFormatter(object):
         option_type = getattr(opt, 'type', None)
         opt_type = getattr(option_type, 'type_name', 'unknown value')
 
+        opt_prefix = ''
+        if (opt.deprecated_for_removal and
+                not opt.help.startswith('DEPRECATED')):
+            opt_prefix = 'DEPRECATED: '
+
         if opt.help:
-            help_text = u'%s (%s)' % (opt.help,
-                                      opt_type)
+            help_text = u'%s%s (%s)' % (opt_prefix,
+                                        opt.help,
+                                        opt_type)
         else:
             help_text = u'(%s)' % opt_type
         lines = self._format_help(help_text)
@@ -189,6 +195,30 @@ class _OptFormatter(object):
                                       for choice in opt.type.choices])
             lines.append('# Allowed values: %s\n' % choices_text)
 
+        try:
+            if opt.mutable:
+                lines.append(
+                    '# Note: This option can be changed without restarting.\n'
+                )
+        except AttributeError as err:
+            # NOTE(dhellmann): keystoneauth defines its own Opt class,
+            # and neutron (at least) returns instances of those
+            # classes instead of oslo_config Opt instances. The new
+            # mutable attribute is the first property where the API
+            # isn't supported in the external class, so we can use
+            # this failure to emit a warning. See
+            # https://bugs.launchpad.net/keystoneauth/+bug/1548433 for
+            # more details.
+            import warnings
+            if not isinstance(opt, cfg.Opt):
+                warnings.warn(
+                    'Incompatible option class for %s (%r): %s' %
+                    (opt.dest, opt.__class__, err),
+                )
+            else:
+                warnings.warn('Failed to fully format sample for %s: %s' %
+                              (opt.dest, err))
+
         for d in opt.deprecated_opts:
             lines.append('# Deprecated group/name - [%s]/%s\n' %
                          (d.group or 'DEFAULT', d.name or opt.dest))
@@ -323,33 +353,51 @@ def on_load_failure_callback(*args, **kwargs):
     raise
 
 
-def _output_opts(f, group, namespaces):
-    f.format_group(group)
-    for (namespace, opts) in sorted(namespaces,
+def _output_opts(f, group, group_data):
+    f.format_group(group_data['object'] or group)
+    for (namespace, opts) in sorted(group_data['namespaces'],
                                     key=operator.itemgetter(0)):
         f.write('\n#\n# From %s\n#\n' % namespace)
         for opt in opts:
             f.write('\n')
-            f.format(opt)
-
-
-def _get_group_name(item):
-    group = item[0]
-    # The keys of the groups dictionary could be an OptGroup. Otherwise the
-    # help text of an OptGroup wouldn't be part of the generated sample
-    # file. It could also be just a plain group name without any further
-    # attributes. That's the reason why we have to differentiate here.
-    return group.name if isinstance(group, cfg.OptGroup) else group
+            try:
+                f.format(opt)
+            except Exception as err:
+                f.write('# Warning: Failed to format sample for %s\n' %
+                        (opt.dest,))
+                f.write('# %s\n' % (err,))
 
 
 def _get_groups(conf_ns):
-    groups = {'DEFAULT': []}
+    """Invert a list of groups by namespace into a dict by group name.
+
+    :param conf_ns: a list of (namespace, [(<group>, [opt_1, opt_2])]) tuples,
+                    such as returned by _list_opts.
+    :returns: {<group_name>, {'object': <group_object>,
+                              'namespaces': [(<namespace>, <opts>)]}}
+
+    <group> may be a string or a group object.
+    <group_name> is always a string.
+    <group_object> will only be set if <group> was a group object in at least
+    one namespace.
+
+    Keying by group_name avoids adding duplicate group names in case a group is
+    added as both an OptGroup and as a str, but still makes the additional
+    OptGroup data available to the output code when possible.
+    """
+    groups = {'DEFAULT': {'object': None, 'namespaces': []}}
     for namespace, listing in conf_ns:
         for group, opts in listing:
             if not opts:
                 continue
-            namespaces = groups.setdefault(group or 'DEFAULT', [])
-            namespaces.append((namespace, opts))
+            group = group if group else 'DEFAULT'
+            is_optgroup = hasattr(group, 'name')
+            group_name = group.name if is_optgroup else group
+            if group_name not in groups:
+                groups[group_name] = {'object': None, 'namespaces': []}
+            if is_optgroup:
+                groups[group_name]['object'] = group
+            groups[group_name]['namespaces'].append((namespace, opts))
     return groups
 
 
@@ -375,9 +423,9 @@ def generate(conf):
     _output_opts(formatter, 'DEFAULT', groups.pop('DEFAULT'))
 
     # output all other config sections with groups in alphabetical order
-    for group, namespaces in sorted(groups.items(), key=_get_group_name):
+    for group, group_data in sorted(groups.items()):
         formatter.write('\n\n')
-        _output_opts(formatter, group, namespaces)
+        _output_opts(formatter, group, group_data)
 
 
 def main(args=None):
